"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateIndexModule = exports.generateTypeScript = exports.validateStatement = exports.codegenTargets = void 0;
const path = require("path");
const Either = require("fp-ts/lib/Either");
const Task = require("fp-ts/lib/Task");
const function_1 = require("fp-ts/lib/function");
const fp_utils_1 = require("./fp-utils");
const pkginfo = require("./pkginfo");
exports.codegenTargets = [
    'pg',
    'postgres',
    'pg-promise',
];
////////////////////////////////////////////////////////////////////////
function validateStatement(stmt) {
    const columnNames = new Set();
    const conflicts = new Set();
    stmt.columns.forEach(({ name }) => {
        if (columnNames.has(name)) {
            conflicts.add(name);
        }
        else {
            columnNames.add(name);
        }
    });
    if (conflicts.size) {
        const dup = [...conflicts.values()].sort().join(', ');
        return Either.left(`Duplicate output columns: ${dup}`);
    }
    return Either.right(stmt);
}
exports.validateStatement = validateStatement;
////////////////////////////////////////////////////////////////////////
function generateTypeScript(types, sourceFileName, target, module, funcName, stmt) {
    const positionalOnly = hasOnlyPositionalParams(stmt);
    return (0, function_1.pipe)(Task.of(typeScriptString(sourceFileName, target, module, funcName, stmt.sql)), Task.ap(funcParams(types, stmt, positionalOnly)), Task.ap(funcReturnType(types, stmt)), Task.ap(Task.of(queryValues(stmt, positionalOnly))), Task.ap(Task.of(outputValue(target, stmt))));
}
exports.generateTypeScript = generateTypeScript;
const typeScriptString = (sourceFileName, target, module, funcName, sql) => (params) => (returnType) => (queryValues) => (outputValue) => generators[target]({
    sourceFileName,
    module,
    funcName,
    sql,
    params,
    returnType,
    queryValues,
    outputValue,
});
const topComment = (sourceFileName) => `\
// Generated by ${pkginfo.name} from ${sourceFileName}.
// Do not edit directly. Instead, edit ${sourceFileName} and re-run ${pkginfo.name}.
`;
const generators = {
    pg({ sourceFileName, module, funcName, params, returnType, sql, queryValues, outputValue, }) {
        return `\
${topComment(sourceFileName)}

import { ClientBase, Pool } from '${module}'

export async function ${funcName}(
  client: ClientBase | Pool${params}
): Promise<${returnType}> {
    const result = await client.query(\`\\
${sql}\`${queryValues})
    return ${outputValue}
}
`;
    },
    postgres({ sourceFileName, module, funcName, params, returnType, sql, queryValues, outputValue, }) {
        return `\
${topComment(sourceFileName)}

import * as postgres from '${module}'

export async function ${funcName}(
  sql: postgres.Sql<{}>${params}
): Promise<${returnType}> {
    const result = await sql.unsafe(\`${sql}\`${queryValues})
    return ${outputValue}
}
`;
    },
    'pg-promise'({ sourceFileName, module, funcName, params, returnType, sql, queryValues, outputValue, }) {
        return `\
${topComment(sourceFileName)}

import * as pgp from '${module}'

export async function ${funcName}(
  client: pgp.IDatabase<any, any>${params}
): Promise<${returnType}> {
    const result = await client.query(\`\\
${sql}\`${queryValues})
    return ${outputValue}
}
`;
    },
};
function hasOnlyPositionalParams(stmt) {
    return stmt.params.every((param) => !!param.name.match(/\$\d+/));
}
function funcReturnType(types, stmt) {
    return (0, function_1.pipe)((0, fp_utils_1.traverseATs)(stmt.columns, columnType(types)), Task.map((columnTypes) => {
        const rowType = '{ ' + columnTypes.join('; ') + ' }';
        switch (stmt.rowCount) {
            case 'zero':
                return 'number'; // return the affected row count
            case 'one':
                return rowType;
            case 'zeroOrOne':
                return `${rowType} | null`;
            case 'many':
                return `Array<${rowType}>`;
        }
    }));
}
const columnType = (types) => (column) => {
    return (0, function_1.pipe)(types.columnType(column), Task.map(({ name, type }) => `${stringLiteral(name)}: ${type}`));
};
function outputValue(target, stmt) {
    let rows, count;
    switch (target) {
        case 'pg':
            rows = 'result.rows';
            count = 'result.rowCount';
            break;
        case 'postgres':
            rows = 'result';
            count = 'result.count';
            break;
        case 'pg-promise':
            rows = 'result';
            count = 'result.length';
    }
    switch (stmt.rowCount) {
        case 'zero':
            return count; // return the affected row count
        case 'one':
            return `${rows}[0]`;
        case 'zeroOrOne':
            return `${count} > 0 ? ${rows}[0] : null`;
        case 'many':
            return rows;
    }
}
function stringLiteral(str) {
    return "'" + str.replace('\\', '\\\\').replace("'", "\\'") + "'";
}
function funcParams(types, stmt, positionalOnly) {
    if (!stmt.params.length) {
        return Task.of('');
    }
    return (0, function_1.pipe)(positionalOnly
        ? positionalFuncParams(types, stmt)
        : namedFuncParams(types, stmt), Task.map((params) => `, ${params}`));
}
function positionalFuncParams(types, stmt) {
    return (0, function_1.pipe)((0, fp_utils_1.traverseATs)(stmt.params, (param) => (0, function_1.pipe)(types.tsType(param.type, param.nullable), Task.map((tsType) => `${param.name}: ${tsType}`))), Task.map((params) => params.join(', ')));
}
function namedFuncParams(types, stmt) {
    return (0, function_1.pipe)(positionalFuncParams(types, stmt), Task.map((params) => `params: { ${params} }`));
}
function queryValues(stmt, positionalOnly) {
    if (!stmt.params.length) {
        return '';
    }
    const prefix = positionalOnly ? '' : 'params.';
    return (', [ ' + stmt.params.map((param) => prefix + param.name).join(', ') + ' ]');
}
function generateIndexModule(dirPath, nestedDirs, modules) {
    const nestedDirsStr = nestedDirs
        .map((dir) => {
        const name = path.relative(dirPath, dir.dirPath);
        return `export * as ${name} from './${name}';`;
    })
        .join('\n');
    const modulesStr = modules
        .map(({ tsFileName, funcName }) => `export { ${funcName} } from './${baseNameWithoutExt(tsFileName)}';`)
        .join('\n');
    return nestedDirsStr + '\n' + modulesStr;
}
exports.generateIndexModule = generateIndexModule;
function baseNameWithoutExt(filePath) {
    return path.parse(filePath).name;
}
