"use strict";
// Modified from https://github.com/jinjor/typed-parser/blob/4228d4871e84ef4417f51058970c2423e077f8af/src/index.ts
Object.defineProperty(exports, "__esModule", { value: true });
exports.bracedSep = exports.braced = exports._ = exports.whitespace = exports.float = exports.int = exports.keyword = exports.symbol = exports.sepUntil1 = exports.sepUntil = exports.manyUntil = exports.sepBy1 = exports.sepBy = exports.many = exports.todo = exports.constant = exports.noop = exports.stringBeforeEndOr = exports.stringUntil = exports.stringBefore = exports.expectString = exports.skip = exports.match = exports.lazy = exports.withContext = exports.attempt = exports.guard = exports.oneOf = exports.end = exports.mapWithRange = exports.map = exports.seq3 = exports.seq2 = exports.seq1 = exports.seqNull = exports.seqConst = exports.seq = exports.run = exports.calcPosition = exports.isParseError = void 0;
/**
 * Judge if an error (or anything else) is a ParseError.
 */
function isParseError(e) {
    return e instanceof ParseErrorImpl;
}
exports.isParseError = isParseError;
class ParseErrorImpl extends Error {
    constructor(source, error) {
        super(error.message);
        this.source = source;
        this.error = error;
        this.positions = new Map();
    }
    get offset() {
        return this.error.offset;
    }
    get position() {
        return this.getPosition(this.offset);
    }
    getPosition(offset) {
        let result = this.positions.get(offset);
        if (result === undefined) {
            result = calcPosition(this.source, offset);
            this.positions.set(offset, result);
        }
        return result;
    }
    explain() {
        let text = '';
        const startPos = this.getPosition(this.error.scope.offset);
        const errorPos = this.position;
        const lines = this.source.split('\n').slice(startPos.row - 1, errorPos.row);
        function appendSubMessages(error, indent) {
            if (indent) {
                text += ' '.repeat(indent) + '- ';
            }
            if (typeof error === 'string') {
                text += error + '\n';
                return;
            }
            if (error.message) {
                const at = indent ? '' : ` at (${errorPos.row}:${errorPos.column})`;
                text += `${error.message}${at}\n`;
                return;
            }
            if (error instanceof ExpectOneOf) {
                const pureExpects = [];
                const others = [];
                for (const e of error.errors) {
                    if (e instanceof Expect) {
                        pureExpects.push(e.name);
                    }
                    else if (e instanceof ExpectOneOf) {
                        pureExpects.push(e.alias);
                    }
                    else {
                        others.push(e);
                    }
                }
                const expectations = 'Expect ' +
                    (pureExpects.length > 1 ? 'one of ' : '') +
                    pureExpects.join(', ') +
                    '\n';
                if (!others.length) {
                    text += expectations;
                }
                else {
                    text += `Multiple parsers were not successful\n`;
                    appendSubMessages(expectations, indent + 2);
                    for (const e of others) {
                        appendSubMessages(e, indent + 2);
                    }
                }
            }
        }
        appendSubMessages(this.error, 0);
        text += '\n';
        for (let r = startPos.row; r <= errorPos.row; r++) {
            const line = lines[r - startPos.row];
            text += `${String(r).padStart(5)}| ${line}\n`;
        }
        text += `${' '.repeat(6 + errorPos.column)}^\n`;
        if (this.error.scope.name) {
            text += 'Context:\n';
        }
        let scope = this.error.scope;
        while (scope && scope.name !== null) {
            const { row, column } = this.getPosition(scope.offset);
            text += `    at ${scope.name} (${row}:${column}) \n`;
            scope = scope.parent;
        }
        return text;
    }
}
function calcPosition(source, offset) {
    const sub = (source + ' ').slice(0, offset + 1);
    const lines = sub.split('\n');
    const row = lines.length;
    const column = lines[lines.length - 1].length;
    return { row, column };
}
exports.calcPosition = calcPosition;
function isFailure(e) {
    return e instanceof AbstractFailure;
}
class AbstractFailure {
    constructor(context) {
        this.scope = context.scope;
        this.offset = context.offset;
    }
}
class Expect extends AbstractFailure {
    constructor(context, what, type) {
        super(context);
        this.what = what;
        this.type = type;
        this.alias = null;
    }
    get message() {
        return `Expect ${this.name}`;
    }
    get name() {
        return this.alias || `${this.type} \`${this.what}\``;
    }
}
class NotFound extends AbstractFailure {
    constructor(context, what, name) {
        super(context);
        this.what = what;
        this.name = name;
    }
    get message() {
        return `${this.name} \`${this.what}\` not found`;
    }
}
class ExpectEnd extends AbstractFailure {
    constructor(context) {
        super(context);
    }
    get message() {
        return `Expect the end of source`;
    }
}
class CustomErr extends AbstractFailure {
    constructor(context, message) {
        super(context);
        this.message = message;
    }
}
class ExpectOneOf extends AbstractFailure {
    constructor(context, errors) {
        super(context);
        this.errors = errors;
        this.alias = null;
    }
    get message() {
        if (this.alias) {
            return `Expected ${this.alias}`;
        }
        return '';
    }
}
class Scope {
    constructor(offset, name, parent) {
        this.offset = offset;
        this.name = name;
        this.parent = parent;
    }
}
class Context {
    constructor() {
        this.offset = 0;
        this.scope = new Scope(0, null);
    }
}
/**
 * Run a parser. It throws ParseError when it fails.
 */
function run(parser, source) {
    const context = new Context();
    const result = parser(source, context);
    if (isFailure(result)) {
        throw new ParseErrorImpl(source, result);
    }
    return result;
}
exports.run = run;
/**
 * Apply given parsers and convert the results to another value.
 */
function seq(...parsers) {
    return (map) => {
        return (source, context) => {
            const values = [];
            for (let i = 0; i < parsers.length; i++) {
                const parser = parsers[i];
                const result = parser(source, context);
                if (isFailure(result)) {
                    return result;
                }
                values[i] = result;
            }
            return map(...values);
        };
    };
}
exports.seq = seq;
/**
 * Apply given parsers and return a constant value
 */
function seqConst(value, ...parsers) {
    return seq(...parsers)(() => value);
}
exports.seqConst = seqConst;
/**
 * Apply given parsers and return null
 */
function seqNull(...parsers) {
    return seqConst(null, ...parsers);
}
exports.seqNull = seqNull;
/**
 * Apply given parsers and return the result of the first parser
 */
function seq1(parserA, ...parsers) {
    return seq(parserA, ...parsers)($1);
}
exports.seq1 = seq1;
/**
 * Apply given parsers and return the result of the second parser
 */
function seq2(parserA, parserB, ...parsers) {
    return seq(parserA, parserB, ...parsers)($2);
}
exports.seq2 = seq2;
/**
 * Apply given parsers and return the result of the third parser
 */
function seq3(parserA, parserB, parserC, ...parsers) {
    return seq(parserA, parserB, parserC, ...parsers)($3);
}
exports.seq3 = seq3;
function $null(..._) {
    return null;
}
function $1(a) {
    return a;
}
function $2(_1, a) {
    return a;
}
function $3(_1, _2, a) {
    return a;
}
/**
 * Apply given parser and convert the result to another value.
 */
function map(f, parser) {
    return (source, context) => {
        const originalOffset = context.offset;
        const result = parser(source, context);
        if (isFailure(result)) {
            return result;
        }
        const result2 = f(result, function toError(message) {
            context.offset = originalOffset;
            return new CustomErr(context, message);
        });
        return result2;
    };
}
exports.map = map;
/**
 * Apply given parser and convert the result to another value along with the source location.
 */
function mapWithRange(f, parser) {
    return (source, context) => {
        const originalOffset = context.offset;
        const start = calcPosition(source, context.offset);
        const result = parser(source, context);
        if (isFailure(result)) {
            return result;
        }
        const end = calcPosition(source, context.offset - 1);
        const result2 = f(result, { start, end }, function toError(message) {
            context.offset = originalOffset;
            return new CustomErr(context, message);
        });
        return result2;
    };
}
exports.mapWithRange = mapWithRange;
/**
 * Only succeeds when position is at the end of the source.
 */
const end = (source, context) => {
    if (source.length !== context.offset) {
        return new ExpectEnd(context);
    }
    return null;
};
exports.end = end;
/**
 * Succeeds when one of given parsers succeeds.
 * Note that no fallback will occur if any one of them consumes even a single character.
 * (See also `attempt()`)
 */
function oneOf(...parsers) {
    return (source, context) => {
        const errors = [];
        const originalOffset = context.offset;
        for (const parser of parsers) {
            const result = parser(source, context);
            if (!(result instanceof AbstractFailure)) {
                return result;
            }
            if (originalOffset === context.offset) {
                errors.push(result);
            }
            else {
                return result;
            }
        }
        return new ExpectOneOf(context, errors);
    };
}
exports.oneOf = oneOf;
/**
 * If the first parser fails, the second will be applied.
 * It looks similar to `oneOf()`, but it will say nothing about the first error when the second fails.
 */
function guard(guarder, parser) {
    return (source, context) => {
        const first = guarder(source, context);
        if (!isFailure(first)) {
            return first;
        }
        return parser(source, context);
    };
}
exports.guard = guard;
/**
 * When the given parser fails, offset will return to the first position
 * it started parsing, even if it consists of multiple parsers.
 * This can be used to force fallback in `oneOf()`, but overuse can lead to poor performance.
 */
function attempt(parser) {
    return (source, context) => {
        const originalOffset = context.offset;
        const result = parser(source, context);
        if (isFailure(result)) {
            context.offset = originalOffset;
            return result;
        }
        return result;
    };
}
exports.attempt = attempt;
/**
 * Add helpful name (ex. array, object, ...) to given parser.
 */
function withContext(name, parser) {
    return (source, context) => {
        const parent = context.scope;
        context.scope = new Scope(context.offset, name, parent);
        const originalOffset = context.offset;
        const result = parser(source, context);
        if (context.offset == originalOffset &&
            (result instanceof Expect || result instanceof ExpectOneOf)) {
            result.alias = name;
        }
        context.scope = parent;
        return result;
    };
}
exports.withContext = withContext;
/**
 * Recursively declared parsers cause infinite loop (and stack overflow).
 * To avoid that, `lazy()` gets the parser only when it is needed.
 */
function lazy(getParser) {
    return (source, context) => {
        const parser = getParser();
        if (!parser) {
            throw new Error('Could not get parser');
        }
        return parser(source, context);
    };
}
exports.lazy = lazy;
/**
 * Get string that matched the regex.
 */
function match(regexString) {
    const regexp = new RegExp(regexString, 'smy');
    return (source, context) => {
        regexp.lastIndex = context.offset;
        const result = regexp.exec(source);
        if (result) {
            const s = result[0];
            context.offset += s.length;
            return s;
        }
        else {
            return new Expect(context, regexString, 'pattern');
        }
    };
}
exports.match = match;
/**
 * Skip a part of source that matched the regex.
 */
function skip(regexString) {
    const regexp = new RegExp(regexString, 'smy');
    return (source, context) => {
        regexp.lastIndex = context.offset;
        if (regexp.test(source)) {
            context.offset = regexp.lastIndex;
        }
        return null;
    };
}
exports.skip = skip;
/**
 * Succeeds if the rest of source starts with the given string.
 * The optional type indicates what that string means.
 */
function expectString(s, type = 'string') {
    return (source, context) => {
        if (source.startsWith(s, context.offset)) {
            context.offset += s.length;
            return null;
        }
        else {
            return new Expect(context, s, type);
        }
    };
}
exports.expectString = expectString;
function _stringUntil(regexString, excludeLast) {
    const regexp = new RegExp(regexString, 'g');
    return (source, context) => {
        regexp.lastIndex = context.offset;
        const result = regexp.exec(source);
        if (!result) {
            return new NotFound(context, regexString, 'pattern');
        }
        const s = source.slice(context.offset, result.index);
        context.offset += s.length + (excludeLast ? 0 : result[0].length);
        return s;
    };
}
/**
 * Gets the string before the given pattern but does not consume the last.
 */
function stringBefore(regexString) {
    return _stringUntil(regexString, true);
}
exports.stringBefore = stringBefore;
/**
 * Get the string before the given pattern and consume the last.
 */
function stringUntil(regexString) {
    return _stringUntil(regexString, false);
}
exports.stringUntil = stringUntil;
/**
 * Gets the string before the given pattern or the end of the source.
 */
function stringBeforeEndOr(regexString) {
    const regexp = new RegExp(regexString, 'g');
    return (source, context) => {
        regexp.lastIndex = context.offset;
        const result = regexp.exec(source);
        let index;
        if (result) {
            index = result.index;
        }
        else {
            index = source.length;
        }
        const s = source.slice(context.offset, index);
        context.offset += s.length;
        return s;
    };
}
exports.stringBeforeEndOr = stringBeforeEndOr;
/**
 * Do nothing
 */
const noop = () => {
    return null;
};
exports.noop = noop;
/**
 * Always succeed and return the constant value.
 */
function constant(t) {
    return () => t;
}
exports.constant = constant;
/**
 * This can be used when the implementation is not done.
 */
function todo(name) {
    throw new Error(`Parser "${name}" is not implemented yet.`);
}
exports.todo = todo;
/**
 * Parse many items while it is possible.
 * If the item parser *partially* succeeds, then the entire parser fails.
 * (See also `attempt()`)
 */
function many(itemParser) {
    return (source, context) => {
        const items = [];
        while (true) {
            const originalOffset = context.offset;
            const result = itemParser(source, context);
            if (originalOffset === context.offset) {
                break;
            }
            if (isFailure(result)) {
                return result;
            }
            items.push(result);
        }
        return items;
    };
}
exports.many = many;
function nextItem(separator, itemParser) {
    return seq(attempt(separator), itemParser)($2);
}
/**
 * Parse zero or more items with given separator.
 */
function sepBy(separator, itemParser) {
    return oneOf(seq(itemParser, many(nextItem(separator, itemParser)))((head, tail) => {
        tail.unshift(head);
        return tail;
    }), constant([]));
}
exports.sepBy = sepBy;
/**
 * Parse one or more items with given separator.
 */
function sepBy1(separator, itemParser) {
    return seq(itemParser, many(nextItem(separator, itemParser)))((head, tail) => {
        tail.unshift(head);
        return tail;
    });
}
exports.sepBy1 = sepBy1;
/**
 * Parse many items until something.
 */
function manyUntil(end, itemParser) {
    return (source, context) => {
        const items = [];
        while (true) {
            if (!isFailure(end(source, context))) {
                break;
            }
            const result = itemParser(source, context);
            if (isFailure(result)) {
                return result;
            }
            items.push(result);
        }
        return items;
    };
}
exports.manyUntil = manyUntil;
/**
 * Parse zero or more items with given separator until something.
 */
function sepUntil(end, separator, itemParser) {
    return guard(map((_) => [], end), sepUntil1(end, separator, itemParser));
}
exports.sepUntil = sepUntil;
/**
 * Parse one or more items with given separator until something.
 */
function sepUntil1(end, separator, itemParser) {
    return seq(itemParser, manyUntil(end, seq(separator, itemParser)($2)))((head, tail) => {
        tail.unshift(head);
        return tail;
    });
}
exports.sepUntil1 = sepUntil1;
/**
 * Expect a symbol like `,`, `"`, `[`, etc.
 */
function symbol(s) {
    return expectString(s, 'symbol');
}
exports.symbol = symbol;
function keyword(s, value) {
    if (value === undefined) {
        return expectString(s, 'keyword');
    }
    return map((_) => value, expectString(s, 'keyword'));
}
exports.keyword = keyword;
/**
 * Parse integer with given regex.
 */
function int(regexString) {
    return map((s, toError) => {
        const n = parseInt(s);
        if (isNaN(n)) {
            return toError(`${s} is not an integer`);
        }
        return n;
    }, match(regexString));
}
exports.int = int;
/**
 * Parse float number with given regex.
 */
function float(regexString) {
    return map((s, toError) => {
        const n = parseFloat(s);
        if (isNaN(n)) {
            return toError(`${s} is not a float`);
        }
        return n;
    }, match(regexString));
}
exports.float = float;
/**
 * Skip whitespace (`\\s*`)
 */
exports.whitespace = skip('\\s*');
/**
 * Alias of `whitespace`
 */
exports._ = exports.whitespace;
/**
 * Parse something between symbols with padding (`whitespace`).
 * (Note: should be renamed to `between`)
 */
function braced(start, end, itemParser) {
    return seq(symbol(start), exports._, itemParser, exports._, symbol(end))($3);
}
exports.braced = braced;
/**
 * Parse something like `[ 1, 2, 3 ]`
 */
function bracedSep(start, end, separator, itemParser) {
    return seq(symbol(start), exports._, sepUntil(seq(exports._, symbol(end))($null), separator, itemParser))($3);
}
exports.bracedSep = bracedSep;
